from setuptools import setup, find_packages

setup(
    name="karma-medeval",
    version="0.1",
    author="Vasanth",
    author_email="vasanth.murukuri@eka.care",
    description="KARMA OpenMedEvalKit",
    url="https://github.com/eka-care/lucid",
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2.7Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
